package anynt.environment;


import java.util.ArrayList;

import anynt.object.Agent;
import anynt.object.Interaction;
import anynt.object.Objeto;
import anynt.space.Observation;


/**
 * Represents an environment where there are objects and agents,
 * in the environment they are available some actions
 * 
 * @author Javier
 * @since 11/04/2011
 */
public abstract class Environment
{
	// Object attributes
	/**
	 * Actions that can be made in the environment
	 */
	protected int[] actions;
	/**
	 * Minimum rewards that can be generated in the environment
	 */
	protected double minReward;
	/**
	 * Maximum rewards that can be generated in the environment
	 */
	protected double maxReward;
	/**
	 * Objects that are in the environment
	 */
	protected ArrayList<Objeto> objects;
	/**
	 * Agents that interacts with the environment
	 */
	protected ArrayList<Agent> agents;

	/**
	 * Auxiliar structure to manage the interactions
	 */
	protected Interaction[] interactions;


	// Constructors
	public Environment()
	{
		actions = null;
		minReward = -1.0;
		maxReward = +1.0;

		objects = new ArrayList<Objeto>();
		agents = new ArrayList<Agent>();
	}

	// Treatment of object attributes
	public final int[] getActions()
	{
		return actions;
	}
	public ArrayList<Objeto> getObjects()
	{
		return objects;
	}
	public ArrayList<Agent> getAgents()
	{
		return agents;
	}
	public final void addAgent(Agent agent)
	{
		objects.add(agent);
		agents.add(agent);
	}
	public final void removeAgent(Agent agent)
	{
		objects.remove(agent);
		agents.remove(agent);
	}

	// Object functions
	public void checkConsistency()
	{
		if (actions == null)
			throw new RuntimeException("Environment actions are not defined");
		if (agents.size() == 0)
			throw new RuntimeException("The Environment has no agents");
		for (Agent agent : agents)
			agent.checkConsistency();
	}
	/**
	 * Reset the environment to its initial definition
	 */
	public void reset()
	{
		for (Agent agent : agents)
			agent.reset();
	}

	/**
	 * Prepares the environment to be used if it is necessary
	 */
	public final void prepare()
	{
		interactions = new Interaction[agents.size()];
		for (int index = 0; index < agents.size(); index++)
			interactions[index] = new Interaction();

		prepareEnvironment();
	}
	protected abstract void prepareEnvironment();
	/**
	 * Send rewards and observations to agents and prepare the new interactions if it is needed
	 */
	public final void beginInteraction(int interactionNumber)
	{
		// Send last rewards to the agents
		sendRewardsToAgents(interactions);
		// Send the observation to the agents
		sendObservationToAgents(interactions);

		prepareInteraction(interactionNumber);
	}
	/**
	 * Send rewards to agents
	 */
	private void sendRewardsToAgents(Interaction[] interactions)
	{
		for (int actualAgent = 0; actualAgent < agents.size(); actualAgent++)
		{
			Agent agent = agents.get(actualAgent);

			// Sends the last Reward to the agent
			double reward = interactions[actualAgent].getReward();
			agent.setReward(reward);
		}
	}
	/**
	 * Send observation to agents
	 */
	private void sendObservationToAgents(Interaction[] interactions)
	{
		Observation observation = takeObservation();
		for (int actualAgent = 0; actualAgent < agents.size(); actualAgent++)
		{
			Agent agent = agents.get(actualAgent);

			// Shows the actual Observation to the agent
			interactions[actualAgent].setObservation(observation);
			agent.setObservation(observation);
		}
	}
	/**
	 * Select the Observation to show to the Agent
	 */
	protected abstract Observation takeObservation();
	/**
	 * Prepare an interaction if its needed
	 */
	protected abstract void prepareInteraction(int interactionNumber);
	/**
	 * Take actions from the agents
	 */
	public final void makeInteraction()
	{
		takeActionsFromAgents(interactions);
	}
	/**
	 * Take actions from agents
	 */
	private void takeActionsFromAgents(Interaction[] interactions)
	{
		for (int actualAgent = 0; actualAgent < agents.size(); actualAgent++)
		{
			Agent agent = agents.get(actualAgent);

			// Asks to the Agent for the next action
			int[] action = agent.getAction();
			long elapsedTime = agent.getBehaviour().getElapsedTime();
			interactions[actualAgent].setAction(action);
			interactions[actualAgent].setElapsedTime(elapsedTime);
		}
	}
	/**
	 * Make agents' actions and actualize the environment
	 */
	public final void finishInteraction()
	{
		makeAgentsActions(interactions);
		actualize();
	}
	/**
	 * Make agents' actions
	 */
	protected void makeAgentsActions(Interaction[] interactions)
	{
		for (int actualAgent = 0; actualAgent < agents.size(); actualAgent++)
		{
			Agent agent = agents.get(actualAgent);
			int[] action = interactions[actualAgent].getAction();
			makeAction(agent, action);
		}
	}
	protected abstract void makeAction(Agent agent, int[] action);
	protected abstract void actualize();
	/**
	 * Finish the exercise
	 */
	public final void finish()
	{
		sendRewardsToAgents(interactions);
	}
}