package anynt.environment;


import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Random;

import anynt.object.Agent;
import anynt.object.EvaluableAgent;
import anynt.object.Interaction;
import anynt.object.RewardsGeneratorAgent;
import anynt.space.Cell;


/**
 * Extends a Lambda Environment to represent a reactive Lambda Environment.
 * In the reactive lambda environment two agents never can share the same cell,
 * so if two agents match on the same cell the last to arrive will throw out the previous agent
 * relocating it to any other cell.
 * If some agents arrive at the same time to a cell, one of them randomly is thrown out.
 * 
 * @author Javier
 * @since 11/04/2011
 */
public class ReactiveLambdaEnvironment extends LambdaEnvironment
{
	/**
	 * Place the agents so they not share any cell
	 */
	protected void initialLocateAgents()
	{
		int numCells = space.getNumberOfCells();
		boolean agentsLocatedCorrectly;

		do
		{
			agentsLocatedCorrectly = true;

			// Locate randomly all the agents
			for (Agent agent : agents)
				locateAgent(agent, new Random().nextInt(numCells) + 1);

			for (Agent agent : agents)
				if (agentShareCellWithEvaluableOrRewardGenerator(agent))
				{
					agentsLocatedCorrectly = false;
					break;
				}
		}
		while (!agentsLocatedCorrectly);
	}
	/**
	 * Relocates reward generators so they not share any cell
	 */
	protected void relocateRewardsGeneratorsAgents()
	{
		int numCells = space.getNumberOfCells();

		locateAgent(goodAgent, new Random().nextInt(numCells) + 1);
		locateAgent(evilAgent, new Random().nextInt(numCells) + 1);

		// Relocates the Good and Evil Agents if they where in the same cell
		while (!rewardsGeneratorIsNotWithAnEvaluableAgentInHisCell(goodAgent)
			|| !rewardsGeneratorIsNotWithAnEvaluableAgentInHisCell(evilAgent)
			|| rewardsGeneratorsShareCell())
		{
			locateAgent(goodAgent, new Random().nextInt(numCells) + 1);
			locateAgent(evilAgent, new Random().nextInt(numCells) + 1);
		}
	}
	/**
	 * Make the actions of the agents ensuring they don't share any cell
	 */
	protected void makeAgentsActions(Interaction[] interactions)
	{
		Hashtable<Agent, Integer> reactiveAgents = new Hashtable<Agent, Integer>();

		// Makes agents actions and gets the reactive agents and in what cell they were
		for (int actualAgent = 0; actualAgent < agents.size(); actualAgent++)
		{
			Agent agent = agents.get(actualAgent);
			if (agent instanceof EvaluableAgent
			 || agent instanceof RewardsGeneratorAgent)
				reactiveAgents.put(agent, agent.getLocation().getNumber());
			int[] action = interactions[actualAgent].getAction();
			this.makeAction(agent, action);
		}
		// Check what reactive agents share cells with others reactive agents and didn't move
		ArrayList<Agent> agentsInConflict = new ArrayList<Agent>();
		for (Agent agent : reactiveAgents.keySet())
			if (agent.getLocation().getNumber() == reactiveAgents.get(agent))
				for (Agent agentInCell : agent.getLocation().getAgents())
					if (agentInCell != agent
					 && reactiveAgents.containsKey(agentInCell))
					{
						agentsInConflict.add(agent);
						break;
					}

		// Check what cells are empty from evaluable agents and reward generators
		ArrayList<Cell> availableCells = new ArrayList<Cell>();
		for (Cell cell : space.getCells())
		{
			boolean available = true;
			for (Agent agent : cell.getAgents())
				if (reactiveAgents.containsKey(agent))
				{
					available = false;
					break;
				}
			if (available)
				availableCells.add(cell);
		}

		// The agents that was in the cell, go to another cell where there isn't any evaluable agent nor a reward generator
		for (Agent agent : agentsInConflict)
		{
			int newCellIndex = new Random().nextInt(availableCells.size());
			Cell newLocation = availableCells.get(newCellIndex);
			availableCells.remove(newCellIndex);
			reactiveAgents.remove(agent);

			locateAgent(agent, newLocation.getNumber());
		}

		// If N (2 or more) agents are still in the same cell N - 1 move out randomly
		for (Agent agent : reactiveAgents.keySet())
		{
			agentsInConflict.clear();
			// Checks how many reactive agents are in this cell
			for (Agent agentInCell : agent.getLocation().getAgents())
				if (reactiveAgents.containsKey(agentInCell))
					agentsInConflict.add(agentInCell);
			// If there is more than 1 reactive agent in the cell we have to relocate randomly
			if (agentsInConflict.size() > 1)
			{
				Agent remainsAgent = agentsInConflict.get(new Random().nextInt(agentsInConflict.size()));
				for (Agent conflictiveAgent : agentsInConflict)
					if (conflictiveAgent != remainsAgent)
					{
						int newCellIndex = new Random().nextInt(availableCells.size());
						Cell newLocation = availableCells.get(newCellIndex);
						availableCells.remove(newCellIndex);

						locateAgent(conflictiveAgent, newLocation.getNumber());
					}
			}
		}
	}


	/**
	 * It checks if in agent's cell there is other agent with which should not share the cell
	 */
	private boolean agentShareCellWithEvaluableOrRewardGenerator(Agent agent)
	{
		Cell cell = agent.getLocation();

		for (Agent agentInCell : cell.getAgents())
			if (agentInCell != agent
			 && (agentInCell instanceof EvaluableAgent
			 || agentInCell instanceof RewardsGeneratorAgent))
			 return true;
		return false;
	}
	/**
	 * It checks if the reward generator agent cell is not with any other agent with which should not share the cell
	 */
	private boolean rewardsGeneratorIsNotWithAnEvaluableAgentInHisCell(RewardsGeneratorAgent rewardsGenerator)
	{
		Collection<Agent> cellAgents = rewardsGenerator.getLocation().getAgents();
		for (Agent agent : cellAgents)
			if (agent instanceof EvaluableAgent)
				return false;
		return true;
	}
	/**
	 * It checks if both rewards agents are in the same cell
	 */
	private boolean rewardsGeneratorsShareCell()
	{
		return goodAgent.getLocation().equals(evilAgent.getLocation());
	}
}