package anynt.object;


import java.util.ArrayList;

import anynt.object.behaviour.AgentBehaviour;
import anynt.object.moveController.IAgentMoveController;
import anynt.space.Observation;


/**
 * Extends an object to represent an agent
 * The agents are also objects
 * 
 * @author Javier
 * @since 30/03/2011
 */
public class Agent extends Objeto
{
	// Object attributes
	/**
	 * Controller to manage what and how many actions will be made
	 */
	private IAgentMoveController moveController;
	/**
	 * Behavior of the agent for each action
	 */
	private AgentBehaviour behaviour;
	/**
	 * Actions that the agent know to do
	 */
	private int[] knownActions;
	/**
	 * History of the interactions that have been happened
	 */
	protected final ArrayList<Interaction> interactionHistory;
	/**
	 * Maximum time that the agent has to make an action
	 */
	private long maximumTimeToAct;
	/**
	 * Maximum number of observations that will be saved in the history
	 * It simulates the memory of the agent
	 * To have an infinite memory set to 0
	 */
	private int maximumObservationsSaved;


	// Constructors
	public Agent(String name)
	{
		super(name);

		interactionHistory = new ArrayList<Interaction>();
	}

	// Treatment of object attributes
	public IAgentMoveController getMoveController()
	{
		return moveController;
	}
	public void setMoveController(IAgentMoveController moveController)
	{
		this.moveController = moveController;
	}
	public AgentBehaviour getBehaviour()
	{
		return behaviour;
	}
	public void setBehaviour(AgentBehaviour behaviour)
	{
		this.behaviour = behaviour;
		behaviour.setAgent(this);
	}
	public int[] getKnownActions()
	{
		return knownActions;
	}
	public void setKnownActions(int[] actions)
	{
		knownActions = actions;
	}
	public ArrayList<Interaction> getInteractionHistory()
	{
		return interactionHistory;
	}
	public Interaction getActualInteraction()
	{
		return interactionHistory.get(interactionHistory.size() - 1);
	}
	public long getMaximumTimeToAct()
	{
		return maximumTimeToAct;
	}
	public void setMaximumTimeToAct(long time)
	{
		maximumTimeToAct = time;
	}
	public void setMaximumObservationsSaved(int number)
	{
		maximumObservationsSaved = number;
		// If there is a limit on the observations to be saved
		if (maximumObservationsSaved != 0)
		{
			// It calculates the boundary to where observations will be saved
			int frontierIndex = interactionHistory.size() - maximumObservationsSaved;
			// Before observations will be forgotten
			for (int index = frontierIndex - 1; index >= 0; index--)
			{
				Interaction interaction = interactionHistory.get(index);
				if (interaction.getObservation() == null)
					break;
				interaction.setObservation(null);
			}
		}
	}

	// Object functions
	public void checkConsistency()
	{
		if (moveController == null)
			throw new RuntimeException("An Agent MoveController is not defined");
		if (behaviour == null)
			throw new RuntimeException("An Agent Behaviour is not defined");
		behaviour.checkConsistency();
		if (knownActions == null)
			throw new RuntimeException("An Agent actions are not defined");
	}
	/**
	 * Reset the agent to his initial definition
	 */
	public void reset()
	{
		location = null;

		behaviour.reset();
		interactionHistory.clear();
	}

	/**
	 * The Agent acts
	 */
	public int[] getAction()
	{
		Interaction actualInteraction = getActualInteraction();

		int[] actions = moveController.makeAction(this, actualInteraction.getObservation());

		actualInteraction.setAction(actions);
		actualInteraction.setElapsedTime(behaviour.getElapsedTime());

		return actions;
	}
	/**
	 * The Agent receives a reward
	 */
	public void setReward(double reward)
	{
		startNewInteraction();

		getActualInteraction().setReward(reward);
	}
	/**
	 * The Agent starts a new Interaction
	 */
	private void startNewInteraction()
	{
		// If there is a limited memory
		if (maximumObservationsSaved != 0)
		{
			// It forgets the oldest observation
			int lastObservationIndex = interactionHistory.size() - maximumObservationsSaved;
			if (lastObservationIndex >= 0)
				interactionHistory.get(lastObservationIndex).setObservation(null);
		}
		interactionHistory.add(new Interaction());
	}
	/**
	 * The Agent receives an observation
	 */
	public void setObservation(Observation observation)
	{
		getActualInteraction().setObservation(observation);
		// It actualizes the behavior of the agent if it has to be actualized
		if (interactionHistory.size() != 1)
			behaviour.actualize();
	}
}