package anynt.object.behaviour;


import java.util.ArrayList;
import java.util.Random;

import distribution.IDistribution;
import distribution.UniformDistribution;

import anynt.environment.LambdaEnvironment;
import anynt.object.RewardsGeneratorAgent;
import anynt.space.Cell;
import anynt.space.Observation;


/**
 * It extends the behavior of an agent to represent the behavior where the agent
 * knows what action is the one that gives the best reward
 * 
 * @author Javier
 * @since 31/03/2011
 */
public class OracleBehaviour extends AgentBehaviour
{
	// Object attributes
	/**
	 * Environment where the agent is interacting
	 */
	private final LambdaEnvironment environment;


	// Constructors
	public OracleBehaviour(LambdaEnvironment environment)
	{
		this.environment = environment;
	}

	// Object overrides
	public OracleBehaviour clone()
	{
		return new OracleBehaviour(environment);
	}

	// AgentBehaviour overrides
	/**
	 * It returns the next action realized by the agent
	 */
	public int nextAction(Observation observation)
	{
		// Actualize the time that the agent spend to make the action
		long maximumTime = this.agent.getMaximumTimeToAct();
		elapsedTime += new Random().nextLong() % (maximumTime + 1);

		Cell location = agent.getLocation();
		double bestReward = Double.NEGATIVE_INFINITY;
		ArrayList<Integer> bestActions = new ArrayList<Integer>();

		// If the rewards are consumed before they are actualized
		if (environment.getRewardsConsumedBeforeRewardersUpdateCellRewards())
			// Check all the actions that can be made
			for (int action : availableActions())
			{
				// Cell where the agent can be moved
				Cell adyacentCell = location.destinationCell(action);
				// If the action improve to best so far, it's the new best action
				if (adyacentCell.getReward() > bestReward)
				{
					bestActions.clear();
					bestActions.add(action);
					bestReward = adyacentCell.getReward();
				}
				// If the action equals to the best so far, it's also a best action
				else if (adyacentCell.getReward() == bestReward)
					bestActions.add(action);
			}
		// If the rewards are consumed after of been actualized
		else
		{
			int decreaseFactor = environment.getDecreaseFactor();
			RewardsGeneratorAgent goodAgent = environment.getGoodAgent();
			RewardsGeneratorAgent evilAgent = environment.getEvilAgent();

			// It picks the cell where good will be
			int[] goodAction = goodAgent.getActualInteraction().getAction();
			Cell goodNewCell = goodAgent.getLocation();
			for (int index = 0; index < goodAction.length; index++)
				goodNewCell = goodNewCell.destinationCell(goodAction[index]);
			// It picks the cell where evil will be
			int[] evilAction = evilAgent.getActualInteraction().getAction();
			Cell evilNewCell = evilAgent.getLocation();
			for (int index = 0; index < evilAction.length; index++)
				evilNewCell = evilNewCell.destinationCell(evilAction[index]);
			// If both agents will collide in the same cell
			if (goodNewCell.equals(evilNewCell))
			{
				// If evil agent is not going to change of cell, then good neither will
				Cell evilCell = evilAgent.getLocation();
				if (evilNewCell.equals(evilCell))
					goodNewCell = goodAgent.getLocation();
			}

			// Check all the actions that can be realized
			for (int action : availableActions())
			{
				// Cell where the agent can be moved
				Cell adyacentCell = location.destinationCell(action);

				// It calculates what will be the reward that will be obtained if the action is made
				double newRewardValue;
				if (adyacentCell.equals(goodNewCell))
					newRewardValue = goodAgent.getDropReward();
				else if (adyacentCell.equals(evilNewCell))
					newRewardValue = evilAgent.getDropReward();
				else if (decreaseFactor == Integer.MAX_VALUE)
					newRewardValue = 0.0;
				else
					newRewardValue = adyacentCell.getReward() / decreaseFactor;

				// If this action improve the best so far, it's the new best action
				if (newRewardValue > bestReward)
				{
					bestActions.clear();
					bestActions.add(action);
					bestReward = newRewardValue;
				}
				// If this action equals the best so far, it's also a best action
				else if (newRewardValue == bestReward)
					bestActions.add(action);
			}
		}

		// From best actions take one randomly
		IDistribution uniformDistribution = new UniformDistribution();
		return bestActions.get(uniformDistribution.getValue(bestActions.size()));
	}
}