package anynt.object.behaviour;


import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Random;
import java.util.TreeMap;

import anynt.object.Agent;
import anynt.space.Cell;
import anynt.space.Observation;
import anynt.space.Space;

import distribution.IDistribution;
import distribution.UniformDistribution;


/**
 * It extends the behavior of an agent to represent the behavior of QLearning
 * 
 * @author Javier
 * @since 31/03/2011
 */
public class QLearningBehaviour extends AgentBehaviour
{
	// Object attributes
	/**
	 * Matrix where the rewards for each state are stored
	 */
	private final TreeMap<String, Double[]> matrix;
	/**
	 * Learning rate
	 */
	private final double learningRate;
	/**
	 * Random exploration degree
	 */
	private final double randomExplorationDegree;
	/**
	 * Factor that tells the percentage of the history that will be take into account
	 */
	private final double discountFactor;
	/**
	 * Initial values in the reward matrix
	 */
	private final double initialQ;

	/**
	 * Agents that QLearning recognizes
	 */
	protected final Collection<Agent> agents;
	/**
	 * Last state
	 */
	private String lastStateDescription;
	/**
	 * Actual state
	 */
	private String stateDescription;
	/**
	 * Last action made
	 */
	private int lastAction;


	// Constructors
	public QLearningBehaviour(Collection<Agent> agents, double learningRate, double randomExplorationDegree, double discountFactor, double initialQ)
	{
		this.agents = agents;

		matrix = new TreeMap<String, Double[]>();
		this.learningRate = learningRate;
		this.randomExplorationDegree = randomExplorationDegree;
		this.discountFactor = discountFactor;
		this.initialQ = initialQ;
	}

	// Object overrides
	public QLearningBehaviour clone()
	{
		return new QLearningBehaviour(agents, learningRate, randomExplorationDegree, discountFactor, initialQ);
	}

	// AgentBehaviour overrides
	/**
	 * Actualiza el comportamiento del agente
	 */
	public void actualize()
	{
		Space space = (Space) agent.getActualInteraction().getObservation().getObservation();
		lastStateDescription = stateDescription;
		stateDescription = getStateDescription(space);

		// Actualiza la recompensa de la ltima accin realizada
		double reward = getReward(lastStateDescription, lastAction);
		reward *= (1 - learningRate);
		reward += learningRate * (1 + agent.getActualInteraction().getReward() + discountFactor * maxQ(stateDescription));
		setReward(lastStateDescription, lastAction, reward);
	}

	// Treatment of object attributes
	private double getReward(String state, int action)
	{
		if (!matrix.containsKey(state))
			addState(state);
		Double[] rewards = matrix.get(state);
		int[] actions = agent.getKnownActions();
		for (int index = 0; index < actions.length; index++)
			if (actions[index] == action)
				return rewards[index];
		return 0.0;
	}
	private void setReward(String state, int action, double reward)
	{
		Double[] rewards = matrix.get(state);
		int[] actions = agent.getKnownActions();
		for (int index = 0; index < actions.length; index++)
			if (actions[index] == action)
			{
				rewards[index] = reward;
				break;
			}
	}
	public double getLearningRate()
	{
		return learningRate;
	}
	public double getRandomExplorationDegree()
	{
		return randomExplorationDegree;
	}
	public double getDiscountFactor()
	{
		return discountFactor;
	}
	public double getInitialQ()
	{
		return initialQ;
	}

	private void addState(String state)
	{
		int actions = agent.getKnownActions().length;
		Double[] rewards = new Double[actions];
		for (int index = 0; index < rewards.length; index++)
			rewards[index] = initialQ;
		matrix.put(state, rewards);
	}
	/**
	 * It picks the best obtained reward on this state
	 */
	private double maxQ(String state)
	{
		if (!matrix.containsKey(state))
			addState(state);

		// It returns the maximum reward on this state
		return Collections.max(Arrays.asList(matrix.get(state)));
	}

	/**
	 * It returns the next action realized by the agent
	 */
	public int nextAction(Observation observation)
	{
		// It actualizes the time that is spend to make the action
		long maximumTime = this.agent.getMaximumTimeToAct();
		elapsedTime += new Random().nextLong() % (maximumTime + 1);

		Space space = (Space) observation.getObservation();
		if (stateDescription == null)
			stateDescription = getStateDescription(space);

		lastAction = policy();
		return lastAction;
	}
	/**
	 * It returns the description of the provided space
	 */
	private String getStateDescription(Space space)
	{
		String description = "";

		// For each cell
		for (Cell cell : space.getCells())
		{
			// Writes the agents that are in the cell
			for (Agent agent : agents)
				if (cell.getAgents().contains(agent))
					description = description.concat("1");
				else
					description = description.concat("0");
			description = description.concat("|");
		}
		description = description.substring(0, description.length() - 1);

		return description;
	}
	/**
	 * QLearning policy, where is decided the action to make
	 */
	private int policy()
	{
		IDistribution uniformDistribution = new UniformDistribution();

		// If the historic is not take into account make randomly an action
		if (randomExplorationDegree == 1.0)
		{
			Collection<Integer> actions = availableActions();
			return (Integer) actions.toArray()[uniformDistribution.getValue(actions.size())];
		}

		// If the state was never seen then add it
		if (!matrix.containsKey(stateDescription))
			addState(stateDescription);

		double max = Double.NEGATIVE_INFINITY;
		// It gets the historic rewards on this state
		Double[] rewards = matrix.get(stateDescription);
		ArrayList<Integer> bestActions = new ArrayList<Integer>();

		// Check all the possible actions that can be made
		for (int action : availableActions())
		{
			// Reward that has been obtained until now make this action
			// The 'exploration degree' is applied s greater amount of actions can be explored
			double adjustedReward = rewards[action] * uniformDistribution.getValue(1.0 - randomExplorationDegree, 1.0);
			// If this action improves the best so far, it's the new best action
			if (adjustedReward > max)
			{
				max = adjustedReward;
				bestActions.clear();
				bestActions.add(action);
			}
			// If this action equals to the best so far, it's too a best action
			else if (adjustedReward == max)
				bestActions.add(action);
		}

		// From the best actions, choose one randomly
		return bestActions.get(uniformDistribution.getValue(bestActions.size()));
	}
}